# mass-build - Easily Build Software Involving a Large Amount of Source Repositories
# Copyright (C) 2012 Ralf Jung <post@ralfj.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import os, git, subprocess, re

'''A VCS must have an "update" method with an optional "mode" parameter taking one of the three values below,
   a "version" method returning a version name (or None),
   and a "newVersions" method which checks for new versions and prints the result to standard output.'''
MODE_FETCH = 0
MODE_REBASE = 1
MODE_RESET = 2

def natural_sort_key(val):
	return [ (int(c) if c.isdigit() else c) for c in re.split('([0-9]+)', val) ]

def get_non_digit_prefix(val):
	return re.match('[^0-9]*', val).group(0)

# Fetch updates from git
class Git:
	def __init__(self, folder, config):
		self.folder = os.path.abspath(folder)
		self.url = config['url']
		self.commit = config['version']

	class _ProgressPrinter(git.remote.RemoteProgress):
		def update(self, op_code, cur_count, max_count=None, message=''):
			print self._cur_line+(" "*30)+"\r",

	def update(self, mode = MODE_REBASE):
		isBranch = (self.commit.startswith('origin/'))
		if isBranch:
			branchname = self.commit[len('origin/'):]
		else:
			branchname = "tag"
		# get us a git repository, and the "origin" remote
		if os.path.exists(self.folder):
			# load existing repo
			repo = git.Repo(self.folder)
			origin = repo.remotes.origin
			origin.config_writer.set_value("url", self.url) # make sure we use the current URL
		else:
			# create a new one
			os.makedirs(self.folder)
			repo = git.Repo.init(self.folder)
			origin = repo.create_remote('origin', self.url)
		origin.fetch(progress=Git._ProgressPrinter()) # download new data
		print " "*80+"\r", # clean the line we are in
		if mode == MODE_FETCH:
			return
		# create/find correct branch
		if branchname in repo.heads:
			branch = repo.heads[branchname]
		else:
			branch = repo.create_head(branchname, self.commit)
			if isBranch: # track remote branch
				branch.set_tracking_branch(origin.refs[branchname])
		# update it to the latest remote commit
		branch.checkout()
		if mode == MODE_RESET:
			repo.head.reset(self.commit, working_tree=True)
		else:
			repo.git.rebase(self.commit)
		# update submodules
		repo.git.submodule("update", "--init", "--recursive", "--rebase")
		# done
		print "...done",
		if repo.head.reference.commit != repo.commit(self.commit):
			print "(keeping local patches around)",
		print

	def version(self):
		repo = git.Repo(self.folder)
		v = repo.git.describe()
		return v[len(get_non_digit_prefix(v)):] # remove the letter prefix from v (so that it starts with a number)

	def checkVersions(self):
		self.update(mode =  MODE_FETCH)
		repo = git.Repo(self.folder)
		# get tag name for current commit, if any
		commit = repo.commit(self.commit)
		commitTag = filter(lambda t: t.commit == commit, repo.tags)
		if not commitTag:
			print "Version is not a tag"
			return
		currentVersion = str(commitTag[0])
		# get sorted list of tag names with the same letter prefix and higher version number
		tags = map(str, repo.tags)
		tags = filter(lambda t: get_non_digit_prefix(t) == get_non_digit_prefix(currentVersion) and natural_sort_key(t) > natural_sort_key(currentVersion), tags)
		if not tags: return
		tags.sort(key = natural_sort_key)
		print "Versions newer than "+self.commit+" available:"
		print tags

# Fetch updates via SVN
class SVN:
	def __init__(self, folder, url):
		self.folder = os.path.abspath(folder)
		self.url = url

	def update(self, mode = MODE_REBASE):
		if mode == MODE_FETCH: raise Exception("Just fetching is not supported with SVN")
		if os.path.exists(self.folder):
			os.chdir(self.folder) # go into repository
			if mode == MODE_RESET: subprocess.check_call(['svn', 'revert', '-R', '.'])
			subprocess.check_call(['svn', 'switch', self.url]) # and update to the URL we got
		else:
			os.makedirs(self.folder) # if even the parent folder does not exist, svn fails
			subprocess.check_call(['svn', 'co', self.url, self.folder]) # just download it
	
	def version(self):
		return None
	
	def checkVersions(self):
		print "Version checking not supporting with SVN"
